
/* Copyright 2012 Evan Miller */
/* All rights reserved. */

#pragma OPENCL EXTENSION cl_khr_fp64 : enable

#ifdef cl_khr_fp64
typedef double   fcl_fp;
typedef double4  fcl_fp4;
typedef double16 fcl_fp16;
typedef double4  fcl_double4;
typedef double16 fcl_double16;
#else
typedef float    fcl_fp;
typedef float4   fcl_fp4;
typedef float16  fcl_fp16;
typedef float4   fcl_double4;
typedef float16  fcl_double16;
#endif

#define FCL_1_OVER_SQRT_2_PI 0.3989422804014326f
#define FCL_1_OVER_SQRT_2_PI_D 0.3989422804014326

#define FCL_DIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX   8.5
#define FCL_DIGAMMA_MAX_ARG_FOR_LINEAR_APPROX       0.00001
#define FCL_DIGAMMA_D1                             -0.5772156649
#define FCL_DIGAMMA_S3                              0.08333333333
#define FCL_DIGAMMA_S4                              0.0083333333333
#define FCL_DIGAMMA_S5                              0.003968253968

#define FCL_TRIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX  5.
#define FCL_TRIGAMMA_MAX_ARG_FOR_LINEAR_APPROX      0.0001
#define FCL_TRIGAMMA_B2                             0.1666666667
#define FCL_TRIGAMMA_B4                            -0.03333333333
#define FCL_TRIGAMMA_B6                             0.02380952381
#define FCL_TRIGAMMA_B8                            -0.03333333333

#define fcl_slice_to_sliver(i) (4*i)

fcl_fp sum4(fcl_fp4 vec);
fcl_fp sum16(fcl_fp16 vec);
fcl_fp dot16(fcl_fp16 vec1, fcl_fp16 vec2);

fcl_fp4 select4(fcl_fp4 c, fcl_fp4 b, int4 a);
fcl_fp16 select16(fcl_fp16 c, fcl_fp16 b, int16 a);

fcl_fp fcl_logistic_cdf(fcl_fp alpha);
fcl_fp fcl_normal_cdf(fcl_fp alpha);

fcl_fp fcl_normal_pdf(fcl_fp alpha);
fcl_fp fcl_inverse_mills_ratio(fcl_fp alpha);

fcl_fp digamma(fcl_fp alpha);
fcl_fp16 digamma16(fcl_fp16 alpha);

fcl_fp trigamma(fcl_fp alpha);
fcl_fp16 trigamma16(fcl_fp16 alpha);

int4 convert_fp4_to_int4(fcl_fp4 res_fp);
int16 convert_fp16_to_int16(fcl_fp16 res_fp);

fcl_fp4 convert_int4_to_fp4(int4 res_i);
fcl_fp16 convert_int16_to_fp16(int16 res_i);

int4 int_array_elements4(__global int *array, int4 idx);
int16 int_array_elements16(__global int *array, int16 idx);
fcl_fp4 array_elements4(__global fcl_fp *array, int4 idx);
fcl_fp16 array_elements16(__global fcl_fp *array, int16 idx);

fcl_fp dot16(fcl_fp16 vec1, fcl_fp16 vec2) {
    fcl_fp result = 0.f;
    result += dot(vec1.lo.lo, vec2.lo.lo);
    result += dot(vec1.lo.hi, vec2.lo.hi);
    result += dot(vec1.hi.lo, vec2.hi.lo);
    result += dot(vec1.hi.hi, vec2.hi.hi);
    return result;
}

fcl_fp4 select4(fcl_fp4 c, fcl_fp4 b, int4 a) {
    fcl_fp4 res_fp;
    res_fp.s0 = a.s0 ? b.s0 : c.s0;
    res_fp.s1 = a.s1 ? b.s1 : c.s1;
    res_fp.s2 = a.s2 ? b.s2 : c.s2;
    res_fp.s3 = a.s3 ? b.s3 : c.s3;
    return res_fp;
}

fcl_fp16 select16(fcl_fp16 c, fcl_fp16 b, int16 a) {
    fcl_fp16 res_fp;
    res_fp.lo.lo = select4(c.lo.lo, b.lo.lo, a.lo.lo);
    res_fp.lo.hi = select4(c.lo.hi, b.lo.hi, a.lo.hi);
    res_fp.hi.lo = select4(c.hi.lo, b.hi.lo, a.hi.lo);
    res_fp.hi.hi = select4(c.hi.hi, b.hi.hi, a.hi.hi);
    return res_fp;
}

fcl_fp sum4(fcl_fp4 vec) {
    return vec.x + vec.y + vec.z + vec.w;
}

fcl_fp sum16(fcl_fp16 vec) {
    return sum4(vec.lo.lo) + sum4(vec.lo.hi) + sum4(vec.hi.lo) + sum4(vec.hi.hi);
}

fcl_fp fcl_logistic_cdf(fcl_fp alpha) {
    return 1.0 / (1.0 + exp(-alpha));
}

fcl_fp fcl_normal_cdf(fcl_fp alpha) {
    return 0.5 * erfc(-alpha * M_SQRT1_2);
}

fcl_fp fcl_normal_pdf(fcl_fp alpha) {
    return FCL_1_OVER_SQRT_2_PI_D * exp(-0.5 * alpha * alpha);
}

fcl_fp fcl_inverse_mills_ratio(fcl_fp alpha) {
    return fcl_normal_pdf(alpha) / fcl_normal_cdf(alpha);
}

fcl_fp digamma(fcl_fp alpha) {
    fcl_fp r;
    fcl_fp value = 0.0;
    fcl_fp y = alpha;

    if (alpha <= FCL_DIGAMMA_MAX_ARG_FOR_LINEAR_APPROX) {
        return FCL_DIGAMMA_D1 - 1.0 / y;
    }
    while (y < FCL_DIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX) {
        value -= 1.0 / y;
        y += 1.0;
    }
    value += log(y) - 0.5 / y;
    r = 1.0 / (y * y);

    return value - r*(FCL_DIGAMMA_S3 - r*(FCL_DIGAMMA_S4 - r*FCL_DIGAMMA_S5));
}

fcl_fp16 digamma16(fcl_fp16 alpha) {
    fcl_fp16 r;
    fcl_fp16 y = alpha;
    fcl_fp16 value = (fcl_fp16)(0.0);
    while (any(y < FCL_DIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX)) {
        value -= 1.0 / y;
        y += 1.0;
    }
    value += log(y) - 0.5 / y;
    r = 1.0 / (y * y);

    return select(value - r*(FCL_DIGAMMA_S3 - r*(FCL_DIGAMMA_S4 - r*FCL_DIGAMMA_S5)), 
            FCL_DIGAMMA_D1 - 1.0 / alpha, alpha <= FCL_DIGAMMA_MAX_ARG_FOR_LINEAR_APPROX);
}

fcl_fp trigamma(fcl_fp alpha) {
    fcl_fp value = 0.0;
    fcl_fp y;
    fcl_fp z = alpha;

    if (alpha <= FCL_TRIGAMMA_MAX_ARG_FOR_LINEAR_APPROX) {
        return 1.0 / (alpha * alpha);
    }

    while (z < FCL_TRIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX) {
        value += 1.0 / (z * z);
        z += 1.0;
    }

    y = 1.0 / (z * z);

    return value + 0.5 * y + ( 1.0
            + y * ( FCL_TRIGAMMA_B2
                + y * ( FCL_TRIGAMMA_B4
                    + y * ( FCL_TRIGAMMA_B6
                        + y * FCL_TRIGAMMA_B8 )))) / z;
}

fcl_fp16 trigamma16(fcl_fp16 alpha) {
    fcl_fp16 value = 0.0;
    fcl_fp16 y;
    fcl_fp16 z = alpha;

    while (any(z < FCL_TRIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX)) {
        value += 1.0 / (z * z);
        z += 1.0;
    }

    y = 1.0 / (z * z);

    return select(value + 0.5 * y + ( 1.0
            + y * ( FCL_TRIGAMMA_B2
                + y * ( FCL_TRIGAMMA_B4
                    + y * ( FCL_TRIGAMMA_B6
                        + y * FCL_TRIGAMMA_B8 )))) / z,
                        1. / (alpha * alpha), alpha <= FCL_TRIGAMMA_MAX_ARG_FOR_LINEAR_APPROX);
}

int4 convert_fp4_to_int4(fcl_fp4 res_fp) {
    int4 res_i;
    res_i.s0 = res_fp.s0;
    res_i.s1 = res_fp.s1;
    res_i.s2 = res_fp.s2;
    res_i.s3 = res_fp.s3;
    return res_i;
}

int16 convert_fp16_to_int16(fcl_fp16 res_fp) {
    int16 res_i;
    res_i.lo.lo = convert_fp4_to_int4(res_fp.lo.lo);
    res_i.lo.hi = convert_fp4_to_int4(res_fp.lo.hi);
    res_i.hi.lo = convert_fp4_to_int4(res_fp.hi.lo);
    res_i.hi.hi = convert_fp4_to_int4(res_fp.hi.hi);
    return res_i;
}

fcl_fp4 convert_int4_to_fp4(int4 res_i) {
    fcl_fp4 res_fp;
    res_fp.s0 = res_i.s0;
    res_fp.s1 = res_i.s1;
    res_fp.s2 = res_i.s2;
    res_fp.s3 = res_i.s3;
    return res_fp;
}

fcl_fp16 convert_int16_to_fp16(int16 res_i) {
    fcl_fp16 res_fp;
    res_fp.lo.lo = convert_int4_to_fp4(res_i.lo.lo);
    res_fp.lo.hi = convert_int4_to_fp4(res_i.lo.hi);
    res_fp.hi.lo = convert_int4_to_fp4(res_i.hi.lo);
    res_fp.hi.hi = convert_int4_to_fp4(res_i.hi.hi);
    return res_fp;
}

int4 int_array_elements4(__global int *array, int4 idx) {
    int4 res_i;
    res_i.s0 = array[idx.s0];
    res_i.s1 = array[idx.s1];
    res_i.s2 = array[idx.s2];
    res_i.s3 = array[idx.s3];
    return res_i;
}

int16 int_array_elements16(__global int *array, int16 idx) {
    int16 res_i;
    res_i.lo.lo = int_array_elements4(array, idx.lo.lo);
    res_i.lo.hi = int_array_elements4(array, idx.lo.hi);
    res_i.hi.lo = int_array_elements4(array, idx.hi.lo);
    res_i.hi.hi = int_array_elements4(array, idx.hi.hi);
    return res_i;
}

__kernel void fcl_compute_panel_means(
        __global fcl_fp4 *data,
        __global fcl_fp4 *fweights,
        int index,
        int slivers,
        __global int *cluster_sizes,
        __global int *cluster_offsets,
        __global fcl_fp *mean_out) {
    int i = get_global_id(0);

    int last = cluster_offsets[i] + cluster_sizes[i];
    int t;
    fcl_fp sum = 0.f;
    fcl_fp total_weight = 0.f;
    for (t=cluster_offsets[i]; t<last; t++) {
        sum += dot(data[index*slivers+t], fweights[t]);
        total_weight += sum4(fweights[t]);
    }
    mean_out[i] = sum / total_weight;
}

__kernel void fcl_compute_panel_real_sizes(
        __global fcl_fp4 *data,
        __global fcl_fp4 *fweights,
        int index,
        int slivers,
        __global int *cluster_sizes,
        __global int *cluster_offsets,
        __global fcl_fp *size_out) {
    int i = get_global_id(0);

    int last = cluster_offsets[i] + cluster_sizes[i];
    int t;
    fcl_fp sum = 0.f;
    for (t=cluster_offsets[i]; t<last; t++) {
        sum += (fweights[t].s0 > 0.f);
        sum += (fweights[t].s1 > 0.f);
        sum += (fweights[t].s2 > 0.f);
        sum += (fweights[t].s3 > 0.f);
    }
    size_out[i] = sum;
}

__kernel void fcl_compute_fixed_effects_prefix_sum(
        __global fcl_fp4 *prefix,
        __global fcl_fp4 *fweights,
        __global int *clusters,
        __global int *panel_sizes,
        __global int *panel_offsets,
        __global fcl_fp *fe_sum_out) 
{
    int i = get_global_id(0);
    int panel_idx = clusters[i];
    fcl_fp result = 0.f;
    int last = panel_offsets[panel_idx] + panel_sizes[panel_idx];
    int t;
    for (t=panel_offsets[panel_idx]; t<last; t++) {
        result += dot(prefix[t], fweights[t]);
    }
    
    fe_sum_out[i] = result;
}

__kernel void fcl_compute_fixed_effects_mean(
        __global fcl_fp4 *data,
        __global fcl_fp4 *fweights,
        __global int *independent_cols,
        int slivers,
        __global int *id2col,
        __global fcl_fp4 *hessian_prefix,
        __global int *clusters,
        __global int *panel_sizes,
        __global int *panel_offsets,
        __global fcl_fp *fe_sum,
        __global fcl_fp *out_fe_work,
        __global fcl_fp *out_fe_mean) 
{
    int k = get_global_id(0);
    int i = get_global_id(1);
    int cluster_count = get_global_size(1);
    int idx = id2col[k];
    int panel_idx = clusters[i];
    __global fcl_fp4 *col = &data[independent_cols[idx]*slivers];
    int t;
    int last = panel_offsets[panel_idx] + panel_sizes[panel_idx];
    fcl_fp result = 0.f;
    for (t=panel_offsets[panel_idx]; t<last; t++) {
        result += dot(fweights[t] * col[t], hessian_prefix[t]);
    }
    out_fe_work[cluster_count * k + i] = result;
    out_fe_mean[cluster_count * k + i] = result / fe_sum[i];
}

__kernel void fcl_compute_fixed_effects_jacobian(
        __global fcl_fp4 *data,
        __global fcl_fp4 *fweights,
        __global int *independent_cols,
        int slivers,
        __global int *id2col,
        __global fcl_fp4 *jacobian_prefix,
        __global fcl_fp *jacobian_out,
        __global int *clusters,
        int cluster_count,
        __global int *panel_sizes,
        __global int *panel_offsets,
        __global fcl_fp *fe_mean) 
{
	int k = get_global_id(0);
    int idx = id2col[k];
    
	fcl_fp result = 0.f;
	__global fcl_fp4 *col = &data[independent_cols[idx]*slivers];
	int i, t;
    for (i=0; i<cluster_count; i++) {
        fcl_fp mean_val = fe_mean[cluster_count * k + i];
        int panel_idx = clusters[i];
        int last = panel_offsets[panel_idx] + panel_sizes[panel_idx];
        for (t=panel_offsets[panel_idx]; t<last; t++) {
            result += dot(jacobian_prefix[t] * fweights[t], col[t] - (fcl_fp4)(mean_val));
        }
    }
    
	jacobian_out[k] = result;
}

__kernel void fcl_compute_fixed_effects_hessian(
        __global fcl_fp4 *data,
        __global fcl_fp4 *fweights,
        __global int *independent_cols,
        int slivers,
        __global int *id2col,
        __global fcl_fp4 *hessian_prefix,
        __global fcl_fp *hessian_out,
        int hessian_cols_count,
        __global int *clusters,
        int cluster_count,
        __global int *panel_sizes,
        __global int *panel_offsets,
        __global fcl_fp *fe_mean
) {
	int k1 = get_global_id(0);
	int k2 = get_global_id(1);

    int idx1 = id2col[k1];
    int idx2 = id2col[k2];
	fcl_fp result = 0.f;
	__global fcl_fp4 *col1 = &data[independent_cols[idx1]*slivers];
	__global fcl_fp4 *col2 = &data[independent_cols[idx2]*slivers];
	int i, t;

    if (k2 > k1)
        return;

    for (i=0; i<cluster_count; i++) {
        fcl_fp mean_val1 = fe_mean[cluster_count * k1 + i];
        fcl_fp mean_val2 = fe_mean[cluster_count * k2 + i];
        int panel_idx = clusters[i];
        int last = panel_offsets[panel_idx] + panel_sizes[panel_idx];
        
        for (t=panel_offsets[panel_idx]; t<last; t++) {
            result += dot(hessian_prefix[t] * fweights[t], 
                    (col1[t] - (fcl_fp4)(mean_val1)) * (col2[t] - (fcl_fp4)(mean_val2)));
        }
    }
    
	hessian_out[k1*hessian_cols_count+k2] = result;
	hessian_out[k2*hessian_cols_count+k1] = result;
}

__kernel void fcl_compute_fixed_effects_delta(
        __global fcl_fp *fe_jacobian_sum,
        __global fcl_fp *fe_hessian_sum,
        __global fcl_fp *fe_mean,
        __global fcl_fp *fe_delta_beta,
        int hessian_cols_count,
        __global fcl_fp *fe_delta_out) 
{
    int i = get_global_id(0);
    int cluster_count = get_global_size(0);
    
    fcl_fp dot_product = 0.f;
    int k;
    for (k=0; k<hessian_cols_count; k++) {
        dot_product += fe_mean[cluster_count*k+i] * fe_delta_beta[k];
    }
        
    fe_delta_out[i] = fe_jacobian_sum[i] / fe_hessian_sum[i] - dot_product;
}

__kernel void fcl_compute_fixed_effects(
                                        __global fcl_fp *data_means,
                                        int dependent_col,
                                        __global int *independent_cols,
                                        int independent_cols_count,
                                        __global fcl_fp *estimates,
                                        __global fcl_fp *fe_out
) {
    int i = get_global_id(0);
    int cluster_count = get_global_size(0);
    fcl_fp fitted = 0.f;
    int k;
    for (k=0; k<independent_cols_count; k++) {
        fitted += estimates[k] * data_means[independent_cols[k]*cluster_count+i];
    }
    fe_out[i] = data_means[dependent_col*cluster_count+i] - fitted;
}

__kernel void fcl_compute_fixed_effects_delta_hessian(
        int cluster_count,
        __global fcl_fp *fe_work,
        __global fcl_fp *fe_mean,
        __global fcl_fp *hessian_out,
        int hessian_cols_count) 
{
    int k1 = get_global_id(0);
    int k2 = get_global_id(1);

    int i;
    fcl_fp result = 0.f;

    for (i=0; i<cluster_count; i++) {
        result += fe_work[cluster_count*k1+i] * fe_mean[cluster_count*k2+i];
    }

    hessian_out[hessian_cols_count*k1+k2] -= result;
}

__kernel void fcl_compute_fixed_effects_delta_jacobian(
        int cluster_count,
        __global fcl_fp *fe_mean,
        __global fcl_fp *jacobian_prefix_sum,
        __global fcl_fp *jacobian_out) 
{
    int k = get_global_id(0);

    int i;
    fcl_fp result = 0.f;
    for (i=0; i<cluster_count; i++) {
        result += (jacobian_prefix_sum[i] * fe_mean[cluster_count*k+i]);
    }

    jacobian_out[k] -= result;
}
